from fancy.descriptors.tests.unit.method_descriptor_factories.fake_method_descriptor_factory import \
    FakeMethodDescriptorFactory


class TestMethodDescriptorFactoryBase:
    def test_singleton(self):
        assert FakeMethodDescriptorFactory() is FakeMethodDescriptorFactory()



from copy import copy
from typing import Callable

from fancy.descriptors.method_descriptor_base import MethodDescriptorBase
from fancy.descriptors.method_descriptor_factories import MethodDescriptorFactoryBase, SimpleMethodDescriptorFactory


class MethodDescriptor(MethodDescriptorBase):
    """
    The MethodDescriptor is a kind of proxy, it forwarding requests to the true instance of the descriptor.
    All method descriptors classes should inherit this class, except classes which this proxy forwarded.
    """
    true_instance: MethodDescriptorBase

    def __init__(self, method: Callable, factory: MethodDescriptorFactoryBase = None):
        """
        :param method: input method
        :param factory: a factory decides which kind of true instance will be initialized.
        """
        if factory is None:
            # default factory
            factory = SimpleMethodDescriptorFactory()
        self.true_instance = factory.create(method)

    def __call__(self, *args, **kwargs):
        return self.true_instance.__call__(*args, **kwargs)

    def __get__(self, instance, owner):
        new_instance = self.clone()
        new_instance.true_instance = self.true_instance.__get__(instance, owner)
        return new_instance

    def clone(self):
        return copy(self)

    @classmethod
    def bind(cls, *args, **kwargs):
        def inner(method: Callable):
            return cls(method, *args, **kwargs)

        return inner



from fancy.descriptors.method_descriptor_base import SimpleMethodDescriptor
from fancy.descriptors.method_descriptor_factories import SimpleMethodDescriptorFactory


class TestMethodDescriptorFactories:

    def test_create(self):
        assert isinstance(SimpleMethodDescriptorFactory().create(lambda x: x), SimpleMethodDescriptor)



from fancy.descriptors.method_descriptor_base import MethodDescriptor, SimpleMethodDescriptor
from fancy.descriptors.tests.unit.method_descriptor_base.fake_descriptor import FakeDescriptor


class FakeMarked:

    @SimpleMethodDescriptor
    def marked_simple_method_descriptor(self):
        return 0

    @MethodDescriptor
    def marked_method_descriptor(self):
        return 1

    @FakeDescriptor.bind(value=2)
    def marked_fake_descriptor(self):
        return 2



from typing import Any, Callable

from fancy.descriptors import MethodDescriptor
from fancy.descriptors.method_descriptor_factories import MethodDescriptorFactoryBase


class FakeDescriptor(MethodDescriptor):
    value: Any

    def __init__(self, method: Callable, value: Any, factory: MethodDescriptorFactoryBase = None):
        super().__init__(method, factory=factory)
        self.value = value

    def get_value(self):
        return self.value



from typing import Callable
from unittest.mock import MagicMock

from fancy.descriptors import MethodDescriptor
from fancy.descriptors.method_descriptor_factories import MethodDescriptorFactoryBase


class FakeMethodDescriptorFactory(MethodDescriptorFactoryBase, MagicMock):
    def create(self, method: Callable) -> MethodDescriptor:
        pass



from threading import Lock


class Singleton:
    _instance = None

    _lock = Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance




from abc import ABC, abstractmethod
from typing import Callable

from fancy.descriptors.method_descriptor_base import MethodDescriptorBase
from fancy.descriptors.patterns import Singleton


class MethodDescriptorFactoryBase(ABC, Singleton):

    @abstractmethod
    def create(self, method: Callable) -> MethodDescriptorBase:
        pass



import inspect
from abc import ABC, abstractmethod


class MethodDescriptorBase(ABC):

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """
        MethodDescriptors must implement __call__ because the class is  a wrapper of a method
        , it makes the class become a Callable
        """
        pass

    @abstractmethod
    def __get__(self, instance, owner):
        pass

    @classmethod
    def get_marked_method(cls, instance):
        """
        Get all marked methods of an instance with this descriptor.
        :param instance:
        :return: methods
        """
        return {
            method_name: method for method_name, method in inspect.getmembers(instance) if isinstance(method, cls)
        }




from copy import copy
from typing import Callable

from fancy.descriptors.method_descriptor_base import MethodDescriptorBase


class SimpleMethodDescriptor(MethodDescriptorBase):
    """
    This class provide a simplest implementation of MethodDescriptorBase
    Any class should not mark methods as this descriptor, except in testing.
    See MethodDescriptor
    """

    def __init__(self, method: Callable):
        if not isinstance(method, Callable):
            raise TypeError('method must be a Callable.')
        self.method = method
        self.instance = None
        self.owner = None

    def __call__(self, *args, **kwargs):
        if self.instance is None:
            raise RuntimeError('method must bind to an instance first.')
        return self._get_method_of_instance()(*args, **kwargs)

    def _get_method_of_instance(self):
        return self.method.__get__(self.instance, self.owner)

    def __get__(self, instance, owner=None):
        new_instance = self.clone()
        new_instance.instance = instance
        new_instance.owner = owner

        return new_instance

    def clone(self):
        return copy(self)




from typing import List, Type
from fancy.descriptors.method_descriptor_base import SimpleMethodDescriptor, MethodDescriptorBase, MethodDescriptor
from fancy.descriptors.tests.unit.method_descriptor_base.fake_descriptor import FakeDescriptor
from fancy.descriptors.tests.unit.method_descriptor_base.fake_marked import FakeMarked


class TestMethodDescriptors:
    fake_marked_instance = FakeMarked()
    expected_returns = {
        "marked_simple_method_descriptor": 0,
        "marked_method_descriptor": 1,
        "marked_fake_descriptor": 2
    }

    def test_simple_method_descriptor(self):
        expected_names = ["marked_simple_method_descriptor"]
        self._assert_method_name_and_returned_value_equals(SimpleMethodDescriptor, expected_names)
        assert isinstance(self.fake_marked_instance.marked_simple_method_descriptor, SimpleMethodDescriptor)

    def test_method_descriptor(self):
        expected_names = ["marked_method_descriptor", "marked_fake_descriptor"]
        self._assert_method_name_and_returned_value_equals(MethodDescriptor, expected_names)
        assert isinstance(self.fake_marked_instance.marked_method_descriptor, MethodDescriptor)

    def test_fake_method_descriptor(self):
        expected_names = ["marked_fake_descriptor"]
        self._assert_method_name_and_returned_value_equals(FakeDescriptor, expected_names)
        assert isinstance(self.fake_marked_instance.marked_fake_descriptor, FakeDescriptor)
        assert self.fake_marked_instance.marked_fake_descriptor.get_value() == 2

    def _assert_method_name_and_returned_value_equals(
            self,
            descriptor_type: Type[MethodDescriptorBase],
            expected_names: List[str]
    ):
        method_dict = descriptor_type.get_marked_method(self.fake_marked_instance)

        assert len(expected_names) == len(method_dict)
        for name in expected_names:
            assert name in method_dict.keys()
            assert method_dict[name]() == self.expected_returns[name]
        return method_dict



from typing import Callable

from fancy.descriptors.method_descriptor_factories import MethodDescriptorFactoryBase
from fancy.descriptors.method_descriptor_base import MethodDescriptorBase, SimpleMethodDescriptor


class SimpleMethodDescriptorFactory(MethodDescriptorFactoryBase):
    def create(self, method: Callable) -> MethodDescriptorBase:
        return SimpleMethodDescriptor(method)


class MethodDescriptorFactories:
    simple = SimpleMethodDescriptorFactory()



__all__ = [
    "MethodDescriptor", "MethodDescriptor", "MethodDescriptorBase", "SimpleMethodDescriptor"
]

from .method_descriptor_base import MethodDescriptorBase
from .simple_method_descriptor import SimpleMethodDescriptor
from .method_descriptor import MethodDescriptor



__all__ = [
    "MethodDescriptorFactoryBase", "SimpleMethodDescriptorFactory", "MethodDescriptorFactories"
]

from .method_descriptor_factory_base import MethodDescriptorFactoryBase
from .method_descriptor_factories import SimpleMethodDescriptorFactory, MethodDescriptorFactories
